<?php

/**
 * \file engine.php
 * \brief engine.php \n \n Ce fichier contient un programme php.
 *
 * Ce programme permet de traiter des demandes de traitements.
 *
 * \author Fabien Marty <fabien.marty@veremes.com>
 */
$properties["id_gtf_engine"] = $_SERVER['argv'][1]; //Identifiant du moteur calculé à partir de la valeur passée en argument
$ErrorLicense = $_SERVER['argv'][2]; //Identifiant du moteur calculé à partir de la valeur passée en argument

require_once ("php_engine_conf.inc");
require_once ("vmlib/BD.class.inc");
require_once ("vmlib/logUtil.inc");
require_once("vmlib/dbUtil.inc");
require_once ("gtf_lib/phpUtility.inc");
require_once("vmlib/phpUtil.inc");

require_once ("properties_engine.inc");
require_once ("engine.sql.inc");
//require_once ("PEAR/Mail.php");
require_once("Ldap.class.inc");

require_once("vmlib/Email.class.inc");
require_once("gtf_lib/gtf_object/Order.class.inc");
require_once("vmlib/error.inc");
require_once 'string.inc';

require_once 'gtf_lib/FmeCloud.class.inc';
require_once 'gtf_lib/FmeServer.class.inc';
//require_once 'aws_lib/AmazonS3.class.inc';
require_once 'gtf_lib/FmeCloudForGtf.class.inc';


loadLang("engines", $properties["language"]);
if ($ErrorLicense == "0") {
    //writeToLog('|INFO |PHP| dump prop '. var_export($properties,true),$properties["engine_log_file"]);

    $sDateDebut = new DateTime(date("Y-m-d H:i:s"));
    $sDateFin = new DateTime(date("Y-m-d H:i:s"));

    //Connexion à la base de données de Données gtf : pas de mot de passe pour le robot
    $oBd = new BD($properties['login_scheduler'], $properties['password_scheduler'], $properties["database"], $properties["server"], $properties["port"], $properties["sgbd"], $properties["page_encoding"]);
    if ($oBd->erreurRencontree) {
        writeToLog(INFO_BASE_CONNECTION_ERROR . $properties["database"], $properties["engine_log_file"]);
    } else {
        writeToLog(INFO_BASE_CONNECTION . $properties["database"], $properties["engine_log_file"]);

        // Données du moteur fme utilisé.
        $aParams = array();
        $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
        $aParams['gtf_engine_id'] = array('value' => $properties['id_gtf_engine'], 'type' => 'number');
        $oPDOResult4 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["getFmeEngine"], $aParams);
        $aFmeEngine = $oBd->ligneSuivante($oPDOResult4);
        
        // Si stockage sur S3 -> fichiers du projet FME dans un répertoire temporaire.
        if ($properties['fileS3Uploader'] === true) {
            $properties["workspace_dir"] = $properties['extract_dir'] . '/workspace';
            if (!(file_exists($properties["workspace_dir"])))
                mkdir($properties["workspace_dir"], 0777, true);
        }
        
        // Variables d'environnement.
        if ($aFmeEngine['fme_engine_type_id'] == 'desktop') {
            //Recherche du Moteur V2 (ExecuteWithParams)
            //Selection du chemin fme
            $aParams = array();
            $aParams['sSchemaGtf'] = array('value' => $properties['schema_gtf'], 'type' => 'schema_name');
            $aParams['gtf_engine_id'] = array('value' => $properties['id_gtf_engine'], 'type' => 'number');
            $oPDOResult = $oBd->executeWithParams($aSql['pgsql']['select_fme_path'], $aParams);

            if ($aRecord = $oBd->ligneSuivante($oPDOResult)) {
                $properties["fme_path"] = $aRecord['fme_path'];
            } else {
                $properties["fme_path"] = '';
            }
            putenv("FME_PATH=" . $properties["fme_path"]);
            putenv("GTF_HOME=" . $properties["workspace_dir"] . "/");
        }

        /*
        // Traitement des demandes en fonction du moteur
        $iNbTraitement = 0;
        $aUserToRefresh = Array();
        //$sSql = $aSql[$properties["sgbd"]]["boucle_1"];
        if (TesteHeureCreuse($properties["heure_creuse_min"], $properties["heure_creuse_max"])) { // On est en heure creuse (0 correspond à "Heure Creuse")
            $iPlageHorraire = 0;
            $sSql = $aSql[$properties["sgbd"]]["boucle_1"];
        } else { // On est PAS en heure creuse
            $iPlageHorraire = 0;
            $sSql = $aSql[$properties["sgbd"]]["boucle_2"];
        }
        $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
        $sSql = str_replace("[iPriorityId]", $iPlageHorraire, $sSql);
        $sSql = str_replace("[iIdGtfEngine]", $properties["id_gtf_engine"], $sSql);
        if ($properties["nbr_order_max"] < 1)
            $properties["nbr_order_max"] = 1;
        $sSql = str_replace("[iNnbrOrderMax]", $properties["nbr_order_max"], $sSql);
        $sEnErreur = 2;
        $oPDOResult = $oBd->execute($sSql);
        //*/
        
        ///*
        // Traitement des demandes en fonction du moteur7
        // Gestion des heures creuses
        $iNbTraitement = 0;
        $aUserToRefresh = Array();
        $aParams = array();

        $iPlageHorraire = 0;
        
        $aParams['sSchemaGtf'] = array('value' => $properties['schema_gtf'], 'type' => 'schema_name');
        $aParams['iPriorityId'] = array('value' => $iPlageHorraire, 'type' => 'number');
        $aParams['iIdGtfEngine'] = array('value' => $properties["id_gtf_engine"], 'type' => 'number');
        if ($properties["nbr_order_max"] < 1)
            $properties["nbr_order_max"] = 1;
        $aParams['iNnbrOrderMax'] = array('value' => $properties["nbr_order_max"], 'type' => 'number');
        $aParams['sTimeZone'] = array('value' => str_replace('UTC', '', $properties["server_timezone"]), 'type' => 'string');

        $sEnErreur = 2;
        
        if (TesteHeureCreuse($properties["heure_creuse_min"], $properties["heure_creuse_max"])) { // On est en heure creuse (0 correspond à "Heure Creuse")
            $oPDOResult = $oBd->executeWithParams($aSql[$properties["sgbd"]]["boucle_1"], $aParams);
        } else { // On est PAS en heure creuse
            $oPDOResult = $oBd->executeWithParams($aSql[$properties["sgbd"]]["boucle_2"], $aParams);
        }
        $bAsync = true;
        $properties['running_jobs_file_path'] = __DIR__ . '/' . 'gtf_engine' . $properties["id_gtf_engine"] . '_jobid.txt';
        // Suppression des traitements asynchrones sur Fme Server dont les demandes ont été supprimées sur GTF.
        //removeFmeServerDeletedJobs();
        // Traitement des demandes.
        while ($aDemande = $oBd->ligneSuivante($oPDOResult)) {
            try {
                //test pour connaitre les nombres de tentatives de chaque demande
                //if ($aDemande['attempt']>=1){    modif par og 26/04/13
                if ($properties["max_attempt"] == '') {
                    $properties["max_attempt"] = 2;
                }
                if ($aDemande['attempt'] >= $properties["max_attempt"] || empty($aDemande['user_id'])) {
                    if ($aDemande['attempt'] >= $properties["max_attempt"])
                        writeToLog(str_replace("[aDemande['order_id']]", $aDemande['order_id'], INFO_TOO_MANY_ORDER_ATTEMPT), $properties["engine_log_file"]);
                    else
                        writeToLog(str_replace("[aDemande['order_id']]", $aDemande['order_id'], INFO_NO_ASSOCIATED_USER), $properties["engine_log_file"]);
                    $iStatut = 4;
                    
                    //Mise à jour de la demande V2
                    $sDateTraitement = gmdate('Y-m-d H:i:s');
                    if (updateOrder($aDemande["order_id"], $iStatut, $sDateTraitement, null, 0)) {
                        writeToLog(INFO_ORDER_UPDATE, $properties["engine_log_file"]);
                        // Demande non traitable
                        sendWebsocketOrderStatusMessage('order_started', $iStatut);
                    }
                    //$oPDOResult2 = $oBd->fermeResultat();
                } else {

                    /*
                      //Vérification des droits de l'utilsateur
                      $sSql = $aSql[$properties["sgbd"]]["getLogin"];
                      $sSql = str_replace("[sSchemaFramework]", $properties["schema_framework"], $sSql);
                      $sSql = str_replace("[iUserId]", $aDemande['user_id'], $sSql);
                      $oPDOResult2 = $oBd->execute($sSql);
                      // */

                    //Vérification des droits de l'utilsateur
                    // Recuperation des login et droits de l'utilisateur en cour
                    $aParams = array();
                    $aParams['sSchemaFramework'] = array('value' => $properties["schema_framework"], 'type' => 'schema_name');
                    $aParams['iUserId'] = array('value' => $aDemande['user_id'], 'type' => 'number');
                    $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["getLogin"], $aParams);

                    $aUser = $oBd->ligneSuivante($oPDOResult2);
                    $sGroupListId = getUserGroupsEngines($aUser["login"], $oBd, $properties["mixed_rights_management"]);
                    // Mise à jour du nombre de tentative dans la base
                    $iTentative = $aDemande['attempt'] + 1;
                    updateOrderAttempt($aDemande['order_id'], $iTentative);
                    // Mise a jour du message
                    $aParams = array();
                    $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                    $aParams['iOrderId'] = array('value' => $aDemande["order_id"], 'type' => 'number');
                    $aParams['status'] = array('value' => 2, 'type' => 'number');
                    $oPDOResult3 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["update_message"], $aParams);
                    //*/
                    if ($oBd->erreurRencontree) {
                        writeToLog(INFO_TEST_REQUEST_ERROR . '(' . $oBd->getBDMessage() . ')', $properties["engine_log_file"]);
                    }
                    // Fin - OG 27/03/2013 for table job
                    $oPDOResult3 = $oBd->fermeResultat();
                    
                    // OG 27/03/2013 for table job
                    $sBeginExecutionDate = Date('Y-m-d H:i:s');
                    $date = new DateTime();
                    $iTimeStampBegin = $date->getTimestamp(); // Pour calcul de la durée
                    // fin OG 27/03/2013 for job
                    // $sContenuMail = "";
                    $iNbTraitement = $iNbTraitement + 1;
                    $aUserToRefresh[$iNbTraitement] = $aDemande['user_id'];
                    $aUserToRefresh[$iNbTraitement] = $aDemande['user_id'];

                    /*
                      //Vérification si l'utilisateur a les droits pour faire une demande de ce traitement.
                      $sSql = $aSql[$properties["sgbd"]]["right_user"];
                      $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
                      $sSql = str_replace("[sGroupListId]", $sGroupListId, $sSql);
                      $oPDOResult2 = $oBd->execute($sSql);
                      // */

                    ///*
                    //Vérification si l'utilisateur a les droits pour faire une demande de ce traitement.
                    $aParams = array();
                    $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                    $aParams['sGroupListId'] = array('value' => str_replace(',', '|', $sGroupListId), 'type' => 'group');
                    $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["right_user"], $aParams);
                    //*/

                    if ($oBd->erreurRencontree) {
                        writeToLog(INFO_REQUEST_ERROR . $sSql, $properties["engine_log_file"]);
                        writeToLog(INFO_PHP_ERROR, $properties["engine_log_file"]);
                        $sMessage = "Erreur SQL.";
                        // $sContenuMail = 'Votre demande de traitement "'.$aDemande['workspace_nom'].'" a échoué. Voici le message d\'erreur renvoyé : <br><font color="#FF0000">'.$sMessage.'</font><br>';
                        $iStatut = $sEnErreur;
                    } else {
                        if ($sGroupListId == "0") {
                            $bAutorisationDemande = true;
                        } else {
                            $bAutorisationDemande = false;
                            while ($aTraitementsAutorise = $oBd->ligneSuivante($oPDOResult2)) {

                                if ($aTraitementsAutorise['workspace_id'] == $aDemande['workspace_id']) {
                                    $bAutorisationDemande = true;
                                }
                            }
                        }
                        if ($bAutorisationDemande == true) {
                            if (class_exists("Traitement.class.inc") == false) {
                                include_once ("Traitement.class.inc");
                            }

                            /*
                              // recuperer le nom du fichier fmw a traiter
                              $sSql = $aSql[$properties["sgbd"]]["select_traitement"];
                              $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
                              $sSql = str_replace("[iWorkspaceId]", $aDemande['workspace_id'], $sSql);
                              $oPDOResult3 = $oBd->execute($sSql);
                              // */

                            ///*
                            // recuperer le nom du fichier fmw a traiter
                            $aParams = array();
                            $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                            $aParams['iWorkspaceId'] = array('value' => $aDemande['workspace_id'], 'type' => 'number');
                            $oPDOResult3 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["select_traitement"], $aParams);
                            //*/

                            $aTraitement = $oBd->ligneSuivante($oPDOResult3);
                            $iEmailTemplateId = $aTraitement['email_template_id'];
                            $oPDOResult3 = $oBd->fermeResultat();
                            // $sSql = $aSql[$properties["sgbd"]]["select_transit_dir"];
                            // $oPDOResult4 = $oBd->execute($sSql);
                            // $aAction = $oBd->ligneSuivante ($oPDOResult4);
                            // $properties["transit_dir"] = $aAction["value"];
                            // $oPDOResult4 = $oBd->fermeResultat();
                            
                            // Si stockage sur S3 -> Téléchargement de tous les fichiers du projet FME.
                            if ($properties['fileS3Uploader'] === true) {
                                if ($iTentative == 1) {
                                    $sWorkspaceDir = $properties["workspace_dir"] . '/' . $aDemande['workspace_id'];
                                    if (file_exists($sWorkspaceDir))
                                        clearDir($sWorkspaceDir);
                                    if (mkdir($sWorkspaceDir . '/fme', 0777, true)) {
                                        $aTree = getDirectoryInfosFromWsDataDir('gtf', 'workspace', $aDemande['workspace_id'], 'fme');
                                        if (!empty($aTree)) {
                                            if (!empty($aTree[0]['content'])) {
                                                foreach($aTree[0]['content'] as $aFile) {
                                                    $sFmwFileContent = getFileContentInWsDataDir("gtf", "workspace", $aDemande['workspace_id'], 'fme', $aFile['filename']);
                                                    file_put_contents($sWorkspaceDir . '/fme/' . $aFile['filename'], $sFmwFileContent);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            
                            //Lancement du traitement correspondant à la demande.
                            $bValidFmeLicence = false;
                            switch($aFmeEngine['fme_engine_type_id']) {
                                // Traitement exécuté par une occurrence de FME Cloud via l'api rest.
                                case 'fme_cloud_for_gtf':
                                    $oFmeCloudForGtf = new FmeCloudForGtf($aFmeEngine['iam_access_key_id'], $aFmeEngine['iam_secret_access_key'], $aFmeEngine['s3_region'], 'u0jtg81u4c', 'te', $properties['gtf_instance_id']);
                                break;
                                // Traitement exécuté par une occurrence de FME Server sur Fme Cloud.
                                case 'fme_cloud':
                                    $oFmeCloud = new FmeCloud($aFmeEngine['server_url'], $aFmeEngine['fme_cloud_api_token']);
                                    $oFmeCloud->sLogFilePath = $properties["engine_log_file"];
                                    $sMessage = '';
                                    // Liste des occurrences de FME Server.
                                    $aFmeServerInstances = $oFmeCloud->serviceRequest('instances');
                                    if (!empty($aFmeServerInstances)) {
                                        foreach($aFmeServerInstances as $oInstance) {
                                            if ($oInstance->name == $aFmeEngine['fme_server_instance_name'])
                                                $oFmeServerInstance = $oInstance;
                                        }
                                        if (empty($oFmeServerInstance)) {
                                            // Aucune instace active -> mise à jour de la demande (En erreur).
                                            writeToLog(INFO_FME_CLOUD_INSTANCE_NOT_FOUND, $properties["engine_log_file"]);
                                            updateOrder($aDemande["order_id"], 2, gmdate('Y-m-d H:i:s'));
                                            sendWebsocketOrderStatusMessage('order_finished_or_error', 2);
                                            // Envoi un mail de résultat.
                                            sendOrderResultEmail($iEmailTemplateId);
                                            writeToLog(INFO_END_ORDER_PROCESSING, $properties["engine_log_file"]);
                                            // Demande suivante.
                                            continue 2;
                                        }
                                        else {
                                            writeToLog(INFO_FME_CLOUD_INSTANCE_STATUS . $oFmeServerInstance->state, $properties["engine_log_file"]);
                                            // Démarrage de l'occurence si elle est en pause.
                                            if ($oFmeServerInstance->state == 'PAUSED') {
                                                $oFmeCloud->startInstance($oFmeServerInstance->id);
                                                writeToLog(INFO_LAUNCHING_FME_CLOUD_INSTANCE, $properties["engine_log_file"]);
                                                // Nombre de tentative = 0 (instance en pause).
                                                updateOrderAttempt($aDemande["order_id"], 0);
                                                // Demande suivante.
                                                continue 2;
                                            }
                                            else if ($oFmeServerInstance->state == 'RUNNING') {
                                                writeToLog(INFO_FME_CLOUD_INSTANCE_RUNNING, $properties["engine_log_file"]);
                                                // Etat de la demande -> en cours.
                                                $oBd->updateLinkedTable($properties["schema_gtf"] . ".order", "order_status_id", "order_id", $aDemande['order_id'], 5);
                                                sendWebsocketOrderStatusMessage('order_started', 5);
                                                if (processFmeServerOrder('https://' . $oFmeServerInstance->public_ip, $aFmeEngine['login'], $aFmeEngine['password']) === false)
                                                    continue 2;
                                                /*
                                                // Stockage Amazon S3 qui contiendra les fichiers sources pour le traitement FME.
                                                $oAmazonS3 = new AmazonS3($aFmeEngine['s3_access_key_id'], $aFmeEngine['s3_secret_access_key'], $aFmeEngine['s3_region']);
                                                $oAmazonS3->sLogFilePath = $properties["engine_log_file"];
                                                $oTraitement->sFmeServerS3Bucket = $aFmeEngine['fme_server_s3_bucket'];
                                                */
                                                // Mise en pause de l'occurence.
                                                //$oFmeCloud->pauseInstance($oFmeServerInstance->id);
                                            }
                                            else {
                                                // Nombre de tentative = 0 (instance en cours de démarrage).
                                                updateOrderAttempt($aDemande["order_id"], 0);
                                                // Demande suivante.
                                                continue 2;
                                            }
                                        }
                                    }
                                    else {
                                        // Aucune instace active -> mise à jour de la demande.
                                        writeToLog(INFO_NO_FME_CLOUD_INSTANCES, $properties["engine_log_file"]);
                                        updateOrder($aDemande["order_id"], 2, gmdate('Y-m-d H:i:s'));
                                        sendWebsocketOrderStatusMessage('order_finished_or_error', 2);
                                        // Envoi un mail de résultat.
                                        sendOrderResultEmail($iEmailTemplateId);
                                        writeToLog(INFO_END_ORDER_PROCESSING, $properties["engine_log_file"]);
                                        // Demande suivante.
                                        continue 2;
                                    }
                                break;
                                // Traitement exécuté par un serveur fme via l'api rest.
                                case 'fme_server':
                                    // Mise à jour de l'état de la demande de traitement (en cours).
                                    $oBd->updateLinkedTable($properties["schema_gtf"] . ".order", "order_status_id", "order_id", $aDemande['order_id'], 5);
                                    sendWebsocketOrderStatusMessage('order_started', 5);
                                    writeToLog(str_replace("[aDemande['order_id']]", $aDemande['order_id'], INFO_ORDER_PROCESSING), $properties["engine_log_file"]);
                                    // Lancement du traitement.
                                    if (processFmeServerOrder($aFmeEngine['server_url'], $aFmeEngine['login'], $aFmeEngine['password']) === false)
                                        continue 2;
                                break;
                                // Traitement exécuté par un fme local.
                                case 'desktop':
                                    $aLicence = getFmeLicence($properties['fme_path']);
                                    $bLicenceIsValid = $aLicence['valid'];
                                    if ($aLicence['valid'] == 1) {
                                        $bValidFmeLicence = true;
                                        $sDateDebut = new DateTime(date("Y-m-d H:i:s"));
                                        $oTraitement = new Traitement($oBd, $aDemande['order_id'], $properties["engine_log_file"], $aDemande['wk_params'], $properties, $aDemande['workspace_id'] . "/fme/" . $aTraitement['fmw_file']);
                                        if (!$oTraitement->bErreur) {
                                            $oTraitement->setUseExternalDbConnection($aFmeEngine['useexternaldbconnection']);
                                            // Mise à jour de l'état de la demande de traitement (en cours).
                                            $oBd->updateLinkedTable($properties["schema_gtf"] . ".order", "order_status_id", "order_id", $aDemande['order_id'], 5);
                                            sendWebsocketOrderStatusMessage('order_started', 5);
                                            writeToLog(str_replace("[aDemande['order_id']]", $aDemande['order_id'], INFO_ORDER_PROCESSING), $properties["engine_log_file"]);
                                            // Maj du pid dans la base.
                                            $sPidFilePath = realpath(dirname($_SERVER['SCRIPT_FILENAME'])  . '/pid_' . $properties["id_gtf_engine"] . '.txt');
                                            if (file_exists($sPidFilePath)) {
                                                $iPid = file_get_contents($sPidFilePath);
                                                if (is_numeric($iPid))
                                                    updateOrderPid($aDemande['order_id'], $iPid);
                                            }
                                            // Lancement du traitement.
                                            $sResultat = $oTraitement->Process();
                                            $sDateFin = new DateTime(date("Y-m-d H:i:s"));
                                            $sLogFme = $oTraitement->sLogFme;
                                        }
                                        //Gestion des erreurs.

                                        if ($oTraitement->bFmeCrash)
                                            $iStatut = 1;
                                        elseif ($oTraitement->bErreur) {
                                            if ($aTraitement["failed_action_id"] != "" && $oTraitement->sSource != "") {
                                                $aSource = explode("/", $oTraitement->sSource);
                                                $sSourceDirectory = $properties["upload_dir"] . $aSource[0];

                                                if ($aTraitement["failed_action_id"] == 1) {
                                                    clearDir($sSourceDirectory);
                                                    writeToLog(INFO_DELETE_DIRECTORY . $sSourceDirectory . '.', $properties["engine_log_file"]);
                                                } elseif ($aTraitement["failed_action_id"] == 2) {
                                                    /*
                                                      $sSql = $aSql[$properties["sgbd"]]["select_user"];
                                                      $sSql = str_replace("[iUserId]", $aDemande['user_id'], $sSql);
                                                      $sSql = str_replace("[sSchemaFramework]", $properties["schema_framework"], $sSql);
                                                      $oPDOResult2 = $oBd->execute($sSql);
                                                      // */
                                                    ///*
                                                    // Selection de l'utilisateur
                                                    $aParams = array();
                                                    $aParams['iUserId'] = array('value' => $aDemande['user_id'], 'type' => 'number');
                                                    $aParams['sSchemaFramework'] = array('value' => $properties["schema_framework"], 'type' => 'schema_name');
                                                    $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["select_user"], $aParams);
                                                    //*/
                                                    if ($oBd->erreurRencontree) {
                                                        writeToLog(INFO_GET_USER_INFO_ERROR . $aDemande['user_id'] . ". (SQL = " . $sSql . ")", $properties["engine_log_file"]);
                                                    } else {
                                                        $aUser = $oBd->ligneSuivante($oPDOResult2);

                                                        $sUser = userToFolder($aUser["name"]);
                                                        if (file_exists($properties["transit_dir"])) {
                                                            if (!file_exists($properties["transit_dir"] . "/failed/")) {
                                                                mkdir($properties["transit_dir"] . "/failed/");
                                                            }
                                                            if (!file_exists($properties["transit_dir"] . "/failed/" . $sUser)) {
                                                                mkdir($properties["transit_dir"] . "/failed/" . $sUser);
                                                            }
                                                            if (!file_exists($properties["transit_dir"] . "/failed/" . $sUser . "/" . $aTraitement["workspace_id"])) {
                                                                mkdir($properties["transit_dir"] . "/failed/" . $sUser . "/" . $aTraitement["workspace_id"]);
                                                            }
                                                            rename($sSourceDirectory, $properties["transit_dir"] . "/failed/" . $sUser . "/" . $aTraitement["workspace_id"] . "/" . $aDemande["order_id"]);
                                                            writeToLog(str_replace('[sSourceDirectory]', $sSourceDirectory, INFO_MOVING_FILE) . $properties["transit_dir"] . "/failed/" . $sUser . "/" . $aTraitement["workspace_id"] . "/" . $aDemande["order_id"] . '.', $properties["engine_log_file"]);
                                                        } else {
                                                            writeToLog(str_replace("[properties['transit_dir']]", $properties['transit_dir'], INFO_DIRECTORY_NOT_FOUND_ENGINE), $properties["engine_log_file"]);
                                                            $iStatut = $sEnErreur;
                                                        }
                                                    }
                                                }
                                            }
                                            $sMessage = $oTraitement->sMessageErreur;
                                            $iStatut = $sEnErreur;
                                            $sResultat = '';
                                            // $sContenuMail = 'Votre demande de traitement "'.$aDemande['workspace_nom'].'" a échoué. Voici le message d\'erreur renvoyé : <br><font color="#FF0000">'.$sMessage.'</font><br>';
                                            // $sContenuMail .= "Contactez l'administrateur pour consulter le fichier de log du robot.<br>";
                                            $sLienLogFme = $sLogFme;
                                        } else {
                                            if ($aTraitement["success_action_id"] != "" && $oTraitement->sSource != "") {
                                                $aSource = explode("/", $oTraitement->sSource);
                                                $sSourceDirectory = $properties["upload_dir"] . $aSource[0];
                                                if ($aTraitement["success_action_id"] == 1) {
                                                    clearDir($sSourceDirectory);
                                                    writeToLog(INFO_DELETE_DIRECTORY . $sSourceDirectory . '.', $properties["engine_log_file"]);
                                                } elseif ($aTraitement["success_action_id"] == 2) {

                                                    /*
                                                      $sSql = $aSql[$properties["sgbd"]]["select_user"];
                                                      $sSql = str_replace("[iUserId]", $aDemande['user_id'], $sSql);
                                                      $sSql = str_replace("[sSchemaFramework]", $properties["schema_framework"], $sSql);
                                                      $oPDOResult2 = $oBd->execute($sSql);
                                                      // */
                                                    ///*
                                                    //Selection de l'utilisateur
                                                    $aParams = array();
                                                    $aParams['iUserId'] = array('value' => $aDemande['user_id'], 'type' => 'number');
                                                    $aParams['sSchemaFramework'] = array('value' => $properties["schema_framework"], 'type' => 'schema_name');
                                                    $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["select_user"], $aParams);
                                                    //*/

                                                    if ($oBd->erreurRencontree) {
                                                        writeToLog(INFO_GET_USER_INFO_ERROR . $aDemande['user_id'] . ". (SQL = " . $sSql . ")", $properties["engine_log_file"]);
                                                    } else {
                                                        $aUser = $oBd->ligneSuivante($oPDOResult2);

                                                        $sUser = userToFolder($aUser["name"]);
                                                        if (file_exists($properties["transit_dir"])) {
                                                            if (!file_exists($properties["transit_dir"] . "/success/")) {
                                                                mkdir($properties["transit_dir"] . "/success/");
                                                            }
                                                            if (!file_exists($properties["transit_dir"] . "/success/" . $sUser)) {
                                                                mkdir($properties["transit_dir"] . "/success/" . $sUser);
                                                            }
                                                            if (!file_exists($properties["transit_dir"] . "/success/" . $sUser . "/" . $aTraitement["workspace_id"])) {
                                                                mkdir($properties["transit_dir"] . "/success/" . $sUser . "/" . $aTraitement["workspace_id"]);
                                                            }
                                                            rename($sSourceDirectory, $properties["transit_dir"] . "/success/" . $sUser . "/" . $aTraitement["workspace_id"] . "/" . $aDemande["order_id"]);
                                                            writeToLog(str_replace('[sSourceDirectory]', $sSourceDirectory, INFO_MOVING_FILE) . $properties["transit_dir"] . "/success/" . $sUser . "/" . $aTraitement["workspace_id"] . "/" . $aDemande["order_id"] . '.', $properties["engine_log_file"]);
                                                        } else {
                                                            writeToLog(str_replace("[properties['transit_dir']]", $properties['transit_dir'], INFO_DIRECTORY_NOT_FOUND_ENGINE), $properties["engine_log_file"]);
                                                            $iStatut = $sEnErreur;
                                                        }
                                                    }
                                                }
                                            }
                                            $sMessage = "";
                                            $iStatut = 3;
                                            $sLienLogFme = $sLogFme;
                                        }
                                        $bDataEncrypt = $oTraitement->bDataEncrypt;
                                        unset($oTraitement);
                                    }
                                break;
                            }
                            // Licence invalide.
                            if (!$bValidFmeLicence) {
                                $iTentative = $iTentative - 1;
                                updateOrderAttempt($aDemande['order_id'], $iTentative);
                                writeToLog(INFO_INVALID_FME_LICENCE_FILE, $properties["engine_log_file"]);
                                writeToLog(INFO_CONTACT_ADMINISTRATOR, $properties["engine_log_file"]);
                                $sMessage = "Invalid FME licence file. ";
                                // $sContenuMail = 'Votre demande de traitement "'.$aDemande['workspace_nom'].'" a échoué. Voici le message d\'erreur renvoyé : <br><font color="#FF0000">'.$sMessage.'</font><br>';
                                $iStatut = 1;
                                $sLienLogFme = 'Non disponible';
                            }

                        } else {
                            writeToLog(INFO_NO_USER_GRANT . $aDemande['workspace_nom'] . ' n\'a pas été défini.', $properties["engine_log_file"]);
                            writeToLog(INFO_CONTACT_ADMINISTRATOR, $properties["engine_log_file"]);
                            $sMessage = "Vous n''avez pas les droits nécessaires pour réaliser ce traitement.";
                            // $sContenuMail = 'Votre demande de traitement "'.$aDemande['workspace_nom'].'" a échoué. Voici le message d\'erreur renvoyé : <br><font color="#FF0000">'.$sMessage.'</font><br>';
                            $iStatut = 6;

                            if (file_exists($sLogFme)) {
                                $sLienLogFme = $sLogFme;
                            } else {
                                $sLienLogFme = 'Non disponible';
                            }
                        }
                    }
                    $oPDOResult2 = $oBd->fermeResultat();

                    //Mise à jour de la demande.

                    if ($iStatut == 3) {
                        /*
                          $sSql = $aSql[$properties["sgbd"]]["update_message"];
                          $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
                          $sSql = str_replace("[iOrderId]", $aDemande["order_id"], $sSql);
                          $oPDOResult2 = $oBd->execute($sSql);
                          // */
                        ///*
                        // Mise a jour du message
                        $aParams = array();
                        $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                        $aParams['iOrderId'] = array('value' => $aDemande["order_id"], 'type' => 'number');
                        $aParams['status'] = array('value' => 1, 'type' => 'number');
                        $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["update_message"], $aParams);
                        //*/
                        if ($oBd->erreurRencontree) {
                            writeToLog(INFO_TEST_REQUEST_ERROR . '(' . $oBd->getBDMessage() . ')', $properties["engine_log_file"]);
                        }
                        // Fin - OG 27/03/2013 for table job
                        $oPDOResult2 = $oBd->fermeResultat();
                    }
                    if ($iStatut == 2 || $iStatut == 4) {

                        /*
                          $sSql = $aSql[$properties["sgbd"]]["delete_message"];
                          $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
                          $sSql = str_replace("[iOrderId]", $aDemande["order_id"], $sSql);
                          $oPDOResult2 = $oBd->execute($sSql);
                          // */
                        ///*

                        // Mise a jour du message
                        $aParams = array();
                        $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                        $aParams['iOrderId'] = array('value' => $aDemande["order_id"], 'type' => 'number');
                        $aParams['status'] = array('value' => 1, 'type' => 'number');
                        $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["update_message"], $aParams);
                        //*/

                        if ($oBd->erreurRencontree) {
                            writeToLog(INFO_TEST_REQUEST_ERROR . '(' . $oBd->getBDMessage() . ')', $properties["engine_log_file"]);
                        }
                        // Fin - OG 27/03/2013 for table job
                        $oPDOResult2 = $oBd->fermeResultat();
                    }
                    $sDateTraitement = gmdate('Y-m-d H:i:s');

                    /*
                    $sSql = $aSql[$properties["sgbd"]]["update_demande"];
                    $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
                    if ($iStatut != 1) {
                        $timeFirst = strtotime(date_format($sDateDebut, 'Y-m-d H:i:s'));
                        $timeSecond = strtotime(date_format($sDateFin, 'Y-m-d H:i:s'));
                        $diff = $timeSecond - $timeFirst;
                        $iNbSec = $diff;
                    } else {
                        $iNbSec = "NULL";
                    }
                    // OG 27/03/2013 for table job
                    $date = new DateTime();
                    $iTimeStampEnd = $date->getTimestamp();
                    $iDelay = $iTimeStampEnd - $iTimeStampBegin;
                    $sSql = str_replace("[iNbSec]", $iDelay, $sSql);
                    $sSql = str_replace("[$iStatut]", $iStatut, $sSql);
                    /* $sSql = str_replace("[sMessage]", $sMessage, $sSql); *//*
                    $sSql = str_replace("[sDateTraitement]", $sDateTraitement, $sSql);
                    $sSql = str_replace("[sResultat]", utf8_encode($sResultat), $sSql);
                    $sSql = str_replace("[iOrderId]", $aDemande["order_id"], $sSql);
                    if ($sLogFme == "") {
                        $sSql = str_replace("'[sLogFme]'", "NULL", $sSql);
                    } else {
                        $sSql = str_replace("[sLogFme]", $sLogFme, $sSql);
                    }

                    $oPDOResult2 = $oBd->execute($sSql);
                    //*/

                    //*
                    // Mise a jour de la demande
                    $aParams = array();
                    $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                    if ($iStatut != 1) {
                        $timeFirst = strtotime(date_format($sDateDebut, 'Y-m-d H:i:s'));
                        $timeSecond = strtotime(date_format($sDateFin, 'Y-m-d H:i:s'));
                        $diff = $timeSecond - $timeFirst;
                        $iNbSec = $diff;
                    } else
                        $iNbSec = "NULL";
                    // OG 27/03/2013 for table job
                    $date = new DateTime();
                    $iTimeStampEnd = $date->getTimestamp();
                    $iDelay = $iTimeStampEnd - $iTimeStampBegin;
                    if (updateOrder($aDemande["order_id"], $iStatut, $sDateTraitement, $sLogFme, $iDelay, $sResultat)) {
                        writeToLog(INFO_ORDER_UPDATE, $properties["engine_log_file"]);
                        // La demande est terminée avec un résultat
                        sendWebsocketOrderStatusMessage('order_finished_with_result', $aDemande['order_status_id']);
                    }
                    if ($iStatut != 1) {
                        /*
                          $sSql = $aSql[$properties["sgbd"]]["insertJob"];
                          $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
                          $sSql = str_replace("[iOrderId]", $aDemande["order_id"], $sSql);
                          $sSql = str_replace("[iWorkspaceId]", $aDemande["workspace_id"], $sSql);
                          $sSql = str_replace("[iUserId]", $aDemande['user_id'], $sSql);
                          $sSql = str_replace("[sBeginExecutionDate]", $sBeginExecutionDate, $sSql);
                          $sSql = str_replace("[iDelay]", $iDelay, $sSql);
                          $sSql = str_replace("[iGtfEngineId]", $properties["id_gtf_engine"], $sSql);
                          $sSql = str_replace("[iStatut]", $iStatut, $sSql);
                          //writeToLog('|INFORM|PHP|Test requête OG : '.$sSql,$properties["engine_log_file"]);
                          $oPDOResult2 = $oBd->execute($sSql);
                          // */
                        ///*
                        // Insert de schemaGtf
                        $aParams = array();
                        $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                        $aParams['iOrderId'] = array('value' => $aDemande["order_id"], 'type' => 'number');
                        $aParams['iWorkspaceId'] = array('value' => $aDemande["workspace_id"], 'type' => 'number');
                        $aParams['iUserId'] = array('value' => $aDemande['user_id'], 'type' => 'number');
                        $aParams['sBeginExecutionDate'] = array('value' => $sBeginExecutionDate, 'type' => 'string');
                        $aParams['iDelay'] = array('value' => $iDelay, 'type' => 'number');
                        $aParams['iGtfEngineId'] = array('value' => $properties["id_gtf_engine"], 'type' => 'number');
                        $aParams['iDelay'] = array('value' => $iDelay, 'type' => 'number');
                        $aParams['iStatut'] = array('value' => $iStatut, 'type' => 'number');
                        //writeToLog('|INFORM|PHP|Test requête OG : '.$sSql,$properties["engine_log_file"]);
                        $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["insertJob"], $aParams);
                        //*/
                        if ($oBd->erreurRencontree) {
                            writeToLog(INFO_TEST_REQUEST_ERROR . '(' . $oBd->getBDMessage() . ')', $properties["engine_log_file"]);
                        }
                        // Fin - OG 27/03/2013 for table job
                        $oPDOResult2 = $oBd->fermeResultat();
                    }
                    // Mise à jour de la demande si resultat est vide
                    if ($sResultat == "") {

                        /*
                          $sSql = $aSql[$properties["sgbd"]]["update_demande_resultat_null"];
                          $sSql = str_replace("[sSchemaGtf]", $properties["schema_gtf"], $sSql);
                          $sSql = str_replace("[iOrderId]", $aDemande["order_id"], $sSql);
                          $oPDOResult2 = $oBd->execute($sSql);
                          // */

                        ///*
                        // Mise a jour de la demande si resultat null
                        $aParams = array();
                        $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                        $aParams['iOrderId'] = array('value' => $aDemande["order_id"], 'type' => 'number');
                        $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["update_demande_resultat_null"], $aParams);
                        //*/

                        if ($oBd->erreurRencontree) {
                            writeToLog(INFO_ORDER_UPDATE_ERROR . ' (' . $oBd->getBDMessage() . ')', $properties["engine_log_file"]);
                        } else {
                            writeToLog(INFO_INFO_ORDER_UPDATE_NULL, $properties["engine_log_file"]);
                            // La demande est terminée où en erreur
                            sendWebsocketOrderStatusMessage('order_finished_or_error', $aDemande['order_status_id']);
                        }
                        $oPDOResult2 = $oBd->fermeResultat();
                    }

                    // Notification par e-mail.
                    $bMailToSend = false;
                    switch (strtoupper(trim($aDemande['email_option_id']))) {
                        case 1:
                            $bMailToSend = True;
                            break;
                        case 2:
                            $bMailToSend = False;
                            break;
                        case 3:
                            if ($iStatut == $sEnErreur) {
                                $bMailToSend = True;
                            }
                            break;
                        case 4:
                            if ($iStatut != $sEnErreur) {
                                $bMailToSend = True;
                            }
                            break;
                        default:
                            $bMailToSend = True;
                    }

                    // Licence invalide.
                    if ($iStatut == 2 && !$bLicenceIsValid) {
                        // Mise à jour de l'état de la demande de traitement (en attente).
                        $oBd->updateLinkedTable($properties["schema_gtf"] . ".order", "order_status_id", "order_id", $aDemande['order_id'], 1);
                        // Licence invalide : 1 mail par jour max.
                        $sToday = date('d/m/Y');
                        $sMailLicenceFile = __DIR__ . '\mail_licence.txt';
                        if (file_exists($sMailLicenceFile))
                            if (file_get_contents($sMailLicenceFile) == $sToday)
                                $bMailToSend = false;
                        file_put_contents($sMailLicenceFile, $sToday);
                        // Mise à jour du nombre de tentative dans la base.
                        updateOrderAttempt($aDemande['order_id'], 0);
                    }

                    // Envoi du mail.
                    if ($bMailToSend == true) {
                        /*
                          $sSql = $aSql[$properties["sgbd"]]["select_user"];
                          $sSql = str_replace("[iUserId]", $aDemande['user_id'], $sSql);
                          $sSql = str_replace("[sSchemaFramework]", $properties["schema_framework"], $sSql);
                          $oPDOResult2 = $oBd->execute($sSql);
                          // */
                        ///*

                        //Mise a jour de la demande si resultat null
                        /*
                        $aParams = array();
                        $aParams['iOrderId'] = array('value' => $aDemande['order_id'], 'type' => 'number');
                        $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                        $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["update_demande_resultat_null"], $aParams);
                        //
                        if ($oBd->erreurRencontree) {
                            writeToLog(INFO_GET_USER_INFO_ERROR . $aDemande['user_id'] . ". (SQL = " . $sSql . ")", $properties["engine_log_file"]);
                        } else {
                        */
                            // $aUser = $oBd->ligneSuivante ($oPDOResult2);
                            // $sContenuMail = "Bonjour ".$aUser["name"].",<br><br><br>".$sContenuMail;
                            // $sContenuMail .= "<br><br><cite><h6>Cet e-mail a été envoyé par un robot, merci de ne pas répondre.</h6></cite><br>";
                            //  
                            // Envoi du mail
                            sendOrderResultEmail($iEmailTemplateId, $sMessage);
                        //}
                        $oPDOResult2 = $oBd->fermeResultat();
                    } else {
                        // pas d'envoi de mail
                        writeToLog(INFO_NO_MAIL_SEND, $properties["engine_log_file"]);
                    }
                } //fin du if sur les tentatives
                //fin de while  
            }
            catch (Exception $e) {
                writeToErrorLog($e->getMessage());
                // Mise a jour du message
                $aParams = array();
                $aParams['sSchemaGtf'] = array('value' => $properties["schema_gtf"], 'type' => 'schema_name');
                $aParams['iOrderId'] = array('value' => $aDemande["order_id"], 'type' => 'number');
                $aParams['status'] = array('value' => 1, 'type' => 'number');
                $oPDOResult2 = $oBd->executeWithParams($aSql[$properties["sgbd"]]["update_message"], $aParams);
                // Mise a jour de la demande.
                updateOrder($aDemande["order_id"], 2, gmdate('Y-m-d H:i:s'), null, 0);
                sendWebsocketOrderStatusMessage('order_finished_or_error', 2);
            }
        }
        if ($iNbTraitement >= 1) {
            //Crée ou modifie le fichier de maj de l'application pour l'utilisateur.
            writeToLog(INFO_END_ORDER_PROCESSING, $properties["engine_log_file"]);
        }
        $oPDOResult = $oBd->fermeResultat();
    }
} else {
    writeToErrorLog(constant($ErrorLicense));
}

//Teste si le traitement est autorisé si la demande est en heure creuse.
//retourne true si le traitement est autorisé, retourne false si le traitement n'est pas autorisé.
function TesteHeureCreuse($sHeureMin, $sHeureMax) {
    $sHeure = date('H');
    $sMinute = date('i');
    $aHeureMin = explode(":", $sHeureMin);
    $aHeureMax = explode(":", $sHeureMax);
    $bMax = false;
    $bMin = false;
    //Test de l'heure minimale.
    if ($sHeure > $aHeureMin[0]) {
        $bMin = true;
    } else {
        if ($sHeure == $aHeureMin[0] && $sMinute >= $aHeureMin[1]) {
            $bMin = true;
        }
    }
    //Test de l'heure maximale. 
    if ($sHeure < $aHeureMax[0]) {
        $bMax = true;
    } else {
        if (($aHeureMax[0] < $aHeureMin[0]) && ($sHeure < 24)) {
            $bMax = true;
        }
        if ($sHeure == $aHeureMax[0] && $sMinute <= $aHeureMax[1]) {
            $bMax = true;
        }
    }
    if ($bMin && $bMax) {
        return true;
    } else {
        return false;
    }
}

/**
 * Mise à jour d'une demande.
 * @param {number} $iOrderId Id de la demande.
 * @param {number} $iStatut Statut de la demande.
 * @param {string} $sExecutionDate Date de la demande.
 * @param {number} $iNbSec Durée de la demande en secondes.
 * @param {string} $sResultUrl Url du résultat de la demande.
 * @param {string} $sLogUrl Url du log de la demande.
 */
function updateOrder($iOrderId, $iStatut, $sExecutionDate, $sLogUrl = null, $iNbSec = null, $sResultUrl = null) {
    // Variables globales.
    global $oBd, $properties, $aSql, $aDemande;
    // Suppression du répertoire temporaire du projet.
    if ($properties['fileS3Uploader'] === true) {
        $aStatus = array(2, 3, 4, 6, 7);
        if (in_array($iStatut, $aStatus)) {
            $sWorkspacePath = $properties['workspace_dir'] . '/' . $aDemande['workspace_id'];
            if (file_exists($sWorkspacePath))
                clearDir($sWorkspacePath);
        }
    }
    // Requete d'update de la demande.
    if (empty($sLogUrl))
        $sLogUrl = null;
    if (empty($sResultUrl))
        $sResultUrl = null;
    $aParams = array();
    $aParams['sSchemaGtf'] = array('value' => $properties['schema_gtf'], 'type' => 'schema_name');
    $aParams['iOrderId'] = array('value' => $iOrderId, 'type' => 'number');
    $aParams['iStatut'] = array('value' => $iStatut, 'type' => 'number');
    $aParams['sDateTraitement'] = array('value' => $sExecutionDate, 'type' => 'string');
    $aParams['iNbSec'] = array('value' => $iNbSec, 'type' => 'number');
    $aParams['sResultat'] = array('value' => $sResultUrl, 'type' => 'string');
    $aParams['sLogFme'] = array('value' => $sLogUrl, 'type' => 'string');
    $oBd->executeWithParams($aSql[$properties["sgbd"]]["update_demande"], $aParams);
    if ($oBd->erreurRencontree) {
        writeToLog(INFO_ORDER_UPDATE_ERROR . '(' . $oBd->getBDMessage() . ')', $properties["engine_log_file"]);
        return false;
    }
    else
        return true;
}

/**
 * Envoi un mail de résultat du traitement FME.
 * @param {number} $iEmailTemplateId Id du template de mail.
 * @param {string} $sMessage Message.
 */
function sendOrderResultEmail($iEmailTemplateId, $sMessage = '') {
    // Variables globales.
    global $oBd, $properties, $aDemande, $oTraitement;
    if (!empty($oTraitement))
        $oTraitement = $oTraitement;
    // Envoi du mail
    $aObjects = array();
    if (!empty($sMessage))
        $aObjects['error'] = $sMessage;
    $aObjects['oOrder'] = new OrderLib($oBd, $aDemande["order_id"], $properties, "v_order");
    $aObjects['oOrder']->formatOrderEmail();
    if (!empty($oTraitement))
        $aObjects['oOrder']->aFields['bDataEncrypt'] = $oTraitement->bDataEncrypt;
    $aObjects['oOrder']->aFields['sApplicationUrl'] = $properties['web_server_name'] . '/' . $properties['application_name'];
    if (empty($iEmailTemplateId))
        $iEmailTemplateId = $properties['default_mail_model'];
    $oEmail = new Email($oBd, $iEmailTemplateId, $properties, $aObjects);
    if (!empty($oEmail->oEmailTemplate->name))
        $oEmail->send();
}

/**
 * Mise à jour du nombre de tentatives de la demande.
 * @param {number} $iOrderId Id de la demande.
 * @param {number} $iAttempt Nombre de tentatives.
 */
function updateOrderAttempt($iOrderId, $iAttempt) {
    // Variables globales.
    global $oBd, $properties;
    // Mise à jour du nombre de tentatives de la demande.
    $oBd->updateLinkedTable($properties["schema_gtf"] . ".order", "attempt", "order_id", $iOrderId, $iAttempt);
}

/**
 * Envoi un message au websocket contenant le statut de la demande.
 * @param {string} $sEvent Nom de l'évènement.
 * @param {number} $iStatut Statut de la demande.
 */
function sendWebsocketOrderStatusMessage($sEvent, $iStatut) {
    // Variables globales.
    global $properties, $aDemande;
    // Envoi du message.
    sendWebsocketMessage($properties['websocket_server'], $properties['websocket_port'], $properties['websocket_alias'], array(
        'action' => 'event',
        'service' => 'GtfEvents',
        'data' => array(
            'event' => $sEvent,
            'order' => array(
                'order_id' => $aDemande['order_id'],
                'order_status_id' => $iStatut,
                'user_id' => $aDemande['user_id'],
                'workspace_id' => $aDemande['workspace_id']
            )
        )
    ));
}

/**
 * Ajoute un traitement asynchrone en cours d'exécution.
 * @param {object} $oTraitement Objet de la classe Traitement.
 * @param {string} $sServerUrl Url to Fme Server.
 * @param {string} $sLogin User login.
 * @param {string} $sPassword User password.
 */
function addFmeServerRunningJob($oTraitement, $sServerUrl, $sLogin, $sPassword) {
    // Variables globales.
    global $properties, $aDemande, $aFmeEngine;
    // Liste des demandes asynchrones en cours.
    $aRunningJobs = getFmeServerRunningJobs();
    // Paramètres de la demande asynchrone à ajouter.
    $aRunningJob = array(
        'order_id' => $aDemande['order_id'],
        'fme_engine_type_id' => $aFmeEngine['fme_engine_type_id'],
        'job_id' => $oTraitement->oJobResult->id,
        'result_directory_name' => $oTraitement->sResultDirectoryName,
        'result_directory_path' => $oTraitement->sResultDirectoryPath,
        'destination_files' => $oTraitement->aJobDestinationFiles,
        'resource' => $oTraitement->sJobResource,
        'path' => $oTraitement->sJobPath,
        'fme_server_url' => $sServerUrl,
        'login' => $sLogin,
        'password' => $sPassword
    );
    $aRunningJobs[] = $aRunningJob;
    // Sauve la liste.
    file_put_contents($properties['running_jobs_file_path'], serialize($aRunningJobs));
    // Mise à jour du job id.
    updateOrderPid($aDemande['order_id'], $oTraitement->oJobResult->id);
}

/**
 * Retourne la liste des traitements asynchrones en cours d'exécution.
 * return array
 */
function getFmeServerRunningJobs() {
    // Variables globales.
    global $properties;
    // Retourne la liste des demandes asynchrones en cours ou un tableau vide.
    if (file_exists($properties['running_jobs_file_path']))
        return unserialize(file_get_contents($properties['running_jobs_file_path']));
    else
        return array();
}

/**
 * Vérification de la présence d'un traitement dans la liste des traitements asynchrones en cours d'exécution.
 * @param {number} $iOrderId Id de la demande en cours.
 * @param {array} $aRunningJobs Tableau contenant la liste des traitements asynchrones en cours d'exécution.
 * return boolean
 */
function isFmeServerJobRunning($iOrderId, $aRunningJobs) {
    foreach ($aRunningJobs as $aRunningJob) {
        if ($aRunningJob['order_id'] == $iOrderId)
            return true;
    }
    return false;
}

/**
 * Retourne les infos d'un traitement asynchrones en cours d'exécution.
 * @param {number} $iOrderId I de la demande en cours.
 * @param {array} $aRunningJobs Tableau contenant la liste des traitements asynchrones en cours d'exécution.
 * return array
 */
function getFmeServerRunningJob($iOrderId, $aRunningJobs) {
    foreach ($aRunningJobs as $aRunningJob) {
        if ($aRunningJob['order_id'] == $iOrderId)
            return $aRunningJob;
    }
    return false;
}

/**
 * Supprime un traitement asynchrone en cours d'exécution.
 * @param {number} $iOrderId Id de la demande en cours.
 */
function removeFmeServerRunningJob($iOrderId, $oFmeServer = null) {
    // Variables globales.
    global $properties;
    // Liste des demandes asynchrones en cours.
    $iJobIndex = null;
    $aRunningJobs = getFmeServerRunningJobs();
    foreach ($aRunningJobs as $iIndex => $aRunningJob) {
        if ($aRunningJob['order_id'] == $iOrderId) {
            $iJobIndex = $iIndex;
            break;
        }
    }
    // Sauve la liste.
    if ($iJobIndex !== null) {
        // Sauve la liste des demandes asynchrones en cours.
        $aRunningJob = array_splice($aRunningJobs, $iJobIndex, 1)[0];
        if (empty($aRunningJobs))
            unlink($properties['running_jobs_file_path']);
        else
            file_put_contents($properties['running_jobs_file_path'], serialize($aRunningJobs));
        // Supprime le job id.
        //updateOrderPid($iOrderId);
        // Ecriture dans les logs.
        writeToLog(INFO_FME_SERVER_ASYNCHRONOUS_JOB_REMOVED_FROM_LIST . " (demande $iOrderId).", $properties["engine_log_file"]);
        // Supprime le traitement sur Fme Server.
        if (is_object($oFmeServer)) {
            if (!empty($aRunningJob['job_id']))
                $oFmeServer->deleteJob($aRunningJob['job_id']);
            // Supprime le répertoire des ressources de la demande sur Fme Server.
            if (!empty($aRunningJob['resource']) && !empty($aRunningJob['path']))
                $oFmeServer->serviceRequest('resources/connections/' . $aRunningJob['resource'] . '/filesys/' . $aRunningJob['path'], 'delete');
        }
    }
}

/**
 * Traitement d'une demande sur Fme Server.
 * @param {string} $sServerUrl Url to Fme Server.
 * @param {string} $sLogin User login.
 * @param {string} $sPassword User password.
 * return boolean
 */
function processFmeServerOrder($sServerUrl, $sLogin, $sPassword) {
    // Variables globales utilisées.
    global $oFmeServer, $properties, $oBd, $aFmeEngine, $aDemande, $aTraitement, $aUser, $iEmailTemplateId, $bAsync, $sEnErreur, $iTentative, $sDateDebut, $sDateFin;
    // Variables globales modifiées.
    global $oTraitement, $sMessage, $bValidFmeLicence, $sResultat, $sLogFme, $iStatut, $sLienLogFme;
    // Paramètres de connexion vers Fme Server pour la demande (si elle est dans la liste des traitements asynchrones en cours).
    $aRunningJobs = getFmeServerRunningJobs();
    $aRunningJob = getFmeServerRunningJob($aDemande['order_id'], $aRunningJobs);
    if ($aRunningJob !== false) {
        $sServerUrl = $aRunningJob['fme_server_url'];
        $sLogin = $aRunningJob['login'];
        $sPassword = $aRunningJob['password'];
    }
    // 
    $oFmeServer = new FmeServer($sServerUrl, $sLogin, $sPassword, 'day', 1);
    $oFmeServer->sLogFilePath = $properties["engine_log_file"];
    $sMessage = '';
    // Statut de la licence Fme Server (pas pour Fme Cloud).
    $oLicenseStatus = '';
    if ($aFmeEngine['fme_engine_type_id'] == 'fme_server')
        $oLicenseStatus = $oFmeServer->serviceRequest('licensing/license/status');
    if ($oLicenseStatus === false)
        return false;
    else if ($aFmeEngine['fme_engine_type_id'] == 'fme_cloud' || (!empty($oLicenseStatus) && $oLicenseStatus->isLicensed === true && $oLicenseStatus->isLicenseExpired === false)) {
        $bValidFmeLicence = true;
        $oTraitement = new Traitement($oBd, $aDemande['order_id'], $properties["engine_log_file"], $aDemande['wk_params'], $properties, $aDemande['workspace_id'] . "/fme/" . $aTraitement['fmw_file']);
        $oTraitement->bFmeServer = true;
        // Demande suivante si le traitement est déja en cours d'exécution.
        if ($aRunningJob !== false) {
            // Chemin et nom du répertoire de résultat déja crée.
            $oTraitement->sResultDirectoryName = $aRunningJob['result_directory_name'];
            $oTraitement->sResultDirectoryPath = $aRunningJob['result_directory_path'];
            // Vérification du statut de la demande.
            $oJobRecord = $oTraitement->getFmeServerJobRecord($aRunningJob['job_id'], $oFmeServer);
            // Si le job id n'existe pas : 2 essais puis suppression de la demande dans la liste des traitements asynchrones en cours -> non traitable.
            if ($oFmeServer->aLastCurlRequestInfo['http_code'] == 404) {
                if ($iTentative >= $properties["max_attempt"])
                    removeFmeServerRunningJob($aDemande['order_id']);
            }
            if ($oJobRecord === false) {
                return false;
            }
            $oTraitement->oJobResult = $oJobRecord;
            $aRunningStatus = array('SUBMITTED', 'QUEUED', 'DELAYED', 'PAUSED', 'IN_PROCESS', 'PULLED');
            $aErrorStatus = array('DELETED', 'ABORTED', 'FME_FAILURE', 'JOB_FAILURE');
            writeToLog(INFO_FME_SERVER_JOB_STATUS . $oJobRecord->status, $properties["engine_log_file"]);
            // En cours de traitement.
            if (in_array($oJobRecord->status, $aRunningStatus)) {
                writeToLog(INFO_FME_SERVER_ASYNCHRONOUS_JOB_IN_PROGRESS, $properties["engine_log_file"]);
                // Nombre de tentative = 0 (traitement en cours d'exécution).
                updateOrderAttempt($aDemande["order_id"], 0);
                return false;
            }
            else {
                // Supprime la demande dans la liste des traitements asynchrones en cours.
                removeFmeServerRunningJob($aDemande['order_id']);
                // Télécharge les fichiers de log (et le résultat si présent).
                $sResultat = $oTraitement->downloadFmeServerJobResult($aRunningJob['job_id'], $oFmeServer, $aRunningJob['destination_files'], $aRunningJob['resource'], $aRunningJob['path']);
                $oDateDebut = date_create($oJobRecord->timeStarted);
                $oDateFin = date_create($oJobRecord->timeFinished);
                // Terminé en erreur.
                if (in_array($oJobRecord->status, $aErrorStatus) || $sResultat === false) {
                    updateOrder($aDemande["order_id"], 2, $oJobRecord->timeStarted, $oTraitement->sLogFme, date_diff($oDateDebut, $oDateFin)->format('%s'));
                    sendWebsocketOrderStatusMessage('order_finished_or_error', 2);
                    // Envoi d'un mail de résultat.
                    sendOrderResultEmail($iEmailTemplateId);
                    // Ecriture dans les log.
                    writeToLog(INFO_TREATMENT_FAILURE, $oTraitement->sRobotLogFile);
                    // Demande suivante.
                    return false;
                }
                else if ($oJobRecord->status == 'SUCCESS') {
                    $iStatut = 3;
                    $sLogFme = $oTraitement->sLogFme;
                }
            }
        }
        else {
            $oTraitement->setUseExternalDbConnection($aFmeEngine['useexternaldbconnection']);
            $sResultat = $oTraitement->submitFmeServerJob($oFmeServer, $aDemande, $aTraitement, $aUser, null, $bAsync);
            if ($bAsync) {
                $iStatut = 5;
                addFmeServerRunningJob($oTraitement, $sServerUrl, $sLogin, $sPassword);
                writeToLog(INFO_FME_SERVER_JOB_ID . $oTraitement->oJobResult->id, $properties["engine_log_file"]);
                // Nombre de tentative = 0 (traitement en cours d'exécution).
                updateOrderAttempt($aDemande["order_id"], 0);
                // Demande suivante.
                return false;
            }
            else {
                // Erreur retournée par FME Server.
                if ($sResultat === false) {
                    updateOrder($aDemande["order_id"], 2, gmdate('Y-m-d H:i:s'), $oTraitement->sLogFme);
                    sendWebsocketOrderStatusMessage('order_finished_or_error', 2);
                    writeToLog(INFO_TREATMENT_FAILURE, $oTraitement->sRobotLogFile);
                    // Envoi d'un mail de résultat.
                    sendOrderResultEmail($iEmailTemplateId);
                    // Demande suivante.
                    return false;
                }
                //$sDateFin = new DateTime(date("Y-m-d H:i:s"));
                $sLogFme = $oTraitement->sLogFme;
            }
        }
        // http 200 : $oFmeServer->aLastCurlRequestInfo['http_code']
        $sDateDebut = $oJobRecord->timeStarted;
        $sDateFin = $oJobRecord->timeFinished;
        if ($oTraitement->oJobResult->status == 'SUCCESS') {
            $iStatut = 3;
            $sLienLogFme = $sLogFme;
        }
        else
            $iStatut = $sEnErreur;
    }
    else {
        // Erreur dans la vérification du statut de la licence FME ou expiration.
        if ($oLicenseStatus->isLicenseExpired)
            $ErrorMessage = INFO_FME_SERVER_EXPIRED_LICENCE;
        else
            $ErrorMessage = INFO_LICENCE_STATUS_VERIFICATION_ERROR;
        writeToLog($ErrorMessage, $properties["engine_log_file"]);
        writeToLog(INFO_CONTACT_ADMINISTRATOR, $properties["engine_log_file"]);
        updateOrder($aDemande["order_id"], 2, gmdate('Y-m-d H:i:s'));
        sendWebsocketOrderStatusMessage('order_finished_or_error', 2);
        // Envoi d'un mail de résultat.
        if ($oLicenseStatus->isLicenseExpired === true)
            $sMessage = 'Invalid FME licence file';
        sendOrderResultEmail($iEmailTemplateId, $sMessage);
        // Demande suivante.
        return false;
    }
}

/**
 * Mise à jour du processus ou job id d'une demande.
 * @param {number} $iOrderId Id de la demande.
 * @param {number} $iPid Id du processus (ou job id) de la demande.
 */
function updateOrderPid($iOrderId, $iPid = null) {
    // Variables globales.
    global $oBd, $properties, $aSql;
    // Requete d'update de la demande.
    $aParams = array();
    $aParams['sSchemaGtf'] = array('value' => $properties['schema_gtf'], 'type' => 'schema_name');
    $aParams['iOrderId'] = array('value' => $iOrderId, 'type' => 'number');
    $aParams['iPid'] = array('value' => $iPid, 'type' => 'number');
    $oBd->executeWithParams($aSql[$properties['sgbd']]['update_order_pid'], $aParams);
    if ($oBd->erreurRencontree)
        writeToLog(INFO_ORDER_PID_UPDATE_ERROR . "(SQL : {$aSql[$properties['sgbd']]['update_order_pid']})", $properties['engine_log_file']);
}

/**
 * Supprime tous les traitements sur Fme Server dont la demande a été supprimée dans GTF.
 */
/*
function removeFmeServerDeletedJobs() {
    // Variables globales.
    global $properties, $oBd, $aSql;
    // Requete d'update de la demande.
    $aParams = array();
    $aParams['sSchemaGtf'] = array('value' => $properties['schema_gtf'], 'type' => 'schema_name');
    $oPDOResult = $oBd->executeWithParams($aSql[$properties['sgbd']]['getRunningOrders'], $aParams);
    if ($oBd->erreurRencontree)
        writeToLog(INFO_RUNNING_JOBS_ERROR . "(SQL : {$aSql[$properties['sgbd']]['getRunningOrders']})", $properties['engine_log_file']);
    else {
        $aOrders = array();
        while ($aOrder = $oBd->ligneSuivante($oPDOResult))
            $aOrders[] = $aOrder['order_id'];
        // Liste des demandes asynchrones en cours.
        $aRunningJobs = getFmeServerRunningJobs();
        foreach ($aRunningJobs as $iIndex => $aRunningJob) {
           if (!in_array($aRunningJob['order_id'], $aOrders)) {
                $oFmeServer = null;
                // Supprime le traitement sur Fme Server.
                if (!empty($aRunningJob['job_id'])) {
                    $oFmeServer = new FmeServer($aRunningJob['fme_server_url'], $aRunningJob['login'], $aRunningJob['password'], 'day', 1);
                    $oFmeServer->sLogFilePath = $properties["engine_log_file"];
                    // Statut de la licence Fme Server (pas pour Fme Cloud).
                    if ($aRunningJob['fme_engine_type_id'] == 'fme_server') {
                        $oLicenseStatus = $oFmeServer->serviceRequest('licensing/license/status');
                        if ($oLicenseStatus === false)
                            $oFmeServer = null;
                        else if (is_object($oLicenseStatus) && ($oLicenseStatus->isLicensed === false || $oLicenseStatus->isLicenseExpired === true))
                            $oFmeServer = null;
                    }
                }
                // Supprime le traitement sur Fme Server.
                removeFmeServerRunningJob($aRunningJob['order_id'], $oFmeServer);
            }
        }
    }
}
*/
?>
